# pylama:ignore=W
from payton.scene.geometry.base import Line, Object
from payton.scene.geometry.cube import Cube
from payton.scene.geometry.cylinder import Cylinder
from payton.scene.geometry.md2 import MD2
from payton.scene.geometry.mesh import Mesh
from payton.scene.geometry.particle import ParticleSystem
from payton.scene.geometry.plane import MatrixPlane, Plane
from payton.scene.geometry.ragdoll import Joint, RagDoll
from payton.scene.geometry.sphere import Sphere
from payton.scene.geometry.wavefront import Wavefront

__all__ = [
    "Object",
    "Line",
    "Mesh",
    "Cylinder",
    "Cube",
    "Plane",
    "MatrixPlane",
    "Sphere",
    "ParticleSystem",
    "MD2",
    "Wavefront",
    "RagDoll",
    "Joint",
]



from typing import List, Union

import numpy as np

from payton.math.vector import Vector3D

DIFF = 0.0000001

Matrix = Union[List[Vector3D], np.ndarray]

IDENTITY_MATRIX: Matrix = [
    [1.0, 0.0, 0.0, 0.0],
    [0.0, 1.0, 0.0, 0.0],
    [0.0, 0.0, 1.0, 0.0],
    [0.0, 0.0, 0.0, 1.0],
]

NP_IDENTITY_MATRIX = np.array(IDENTITY_MATRIX, dtype=np.float32)


def matrix_to_position_and_quaternion(
    matrix: Matrix,
) -> tuple[list[float], tuple[float, float, float, float]]:
    """Convert a 4x4 transformation matrix to position and quaternion.

    Args:
        matrix (list of list): A 4x4 transformation matrix.

    Returns:
        tuple: (position, quaternion)
    """
    # Ensure the input is in the form of a numpy array
    matrix = np.array(matrix)

    # Extract the position (translation) part
    position = matrix[3, :3]

    # Extract the rotation part
    rotation_matrix = matrix[:3, :3]
    rotation_matrix = rotation_matrix.T

    # Calculate quaternion components
    trace = rotation_matrix[0, 0] + rotation_matrix[1, 1] + rotation_matrix[2, 2]

    if trace > 0:
        s = 0.5 / np.sqrt(trace + 1.0)
        w = 0.25 / s
        x = (rotation_matrix[2, 1] - rotation_matrix[1, 2]) * s
        y = (rotation_matrix[0, 2] - rotation_matrix[2, 0]) * s
        z = (rotation_matrix[1, 0] - rotation_matrix[0, 1]) * s
    else:
        if (
            rotation_matrix[0, 0] > rotation_matrix[1, 1]
            and rotation_matrix[0, 0] > rotation_matrix[2, 2]
        ):
            s = 2.0 * np.sqrt(
                1.0
                + rotation_matrix[0, 0]
                - rotation_matrix[1, 1]
                - rotation_matrix[2, 2]
            )
            w = (rotation_matrix[2, 1] - rotation_matrix[1, 2]) / s
            x = 0.25 * s
            y = (rotation_matrix[0, 1] + rotation_matrix[1, 0]) / s
            z = (rotation_matrix[0, 2] + rotation_matrix[2, 0]) / s
        elif rotation_matrix[1, 1] > rotation_matrix[2, 2]:
            s = 2.0 * np.sqrt(
                1.0
                + rotation_matrix[1, 1]
                - rotation_matrix[0, 0]
                - rotation_matrix[2, 2]
            )
            w = (rotation_matrix[0, 2] - rotation_matrix[2, 0]) / s
            x = (rotation_matrix[0, 1] + rotation_matrix[1, 0]) / s
            y = 0.25 * s
            z = (rotation_matrix[1, 2] + rotation_matrix[2, 1]) / s
        else:
            s = 2.0 * np.sqrt(
                1.0
                + rotation_matrix[2, 2]
                - rotation_matrix[0, 0]
                - rotation_matrix[1, 1]
            )
            w = (rotation_matrix[1, 0] - rotation_matrix[0, 1]) / s
            x = (rotation_matrix[0, 2] + rotation_matrix[2, 0]) / s
            y = (rotation_matrix[1, 2] + rotation_matrix[2, 1]) / s
            z = 0.25 * s

    quaternion = (x, y, z, w)

    return position.tolist(), quaternion



import logging
from typing import Any

try:
    import gi

    gi.require_version('Gtk', '3.0')
    from gi.repository import Gdk

    _GTK_SUPPORTED = True
except ModuleNotFoundError:
    _GTK_SUPPORTED = False
from payton.scene.camera import BUTTON_LEFT, BUTTON_MIDDLE, BUTTON_RIGHT


class GTKController:
    """GTK controller for eventbox."""

    def __init__(self, scene: Any) -> None:
        """Initialize GTK controller."""
        self.shift_down = False
        self.ctrl_down = False
        self.scene = scene
        self.left_button_down = False
        self.right_button_down = False
        self.middle_button_down = False
        self.xrel = None
        self.yrel = None

    def keyboard_release(self, event: Any) -> None:
        """Release keyboard."""
        keyval = event.keyval
        if keyval == Gdk.KEY_Shift_L:
            self.shift_down = False
        if keyval == Gdk.KEY_Control_L:
            self.shift_down = False

    def button_press(self, event: Any) -> None:
        """Press mouse button."""
        if event.button == 1:
            self.left_button_down = True
            camera = self.scene.active_camera
            mx, my = event.x, event.y
            eye, ray_dir = camera.screen_to_world(mx, my, self.scene.window_width, self.scene.window_height)
            if callable(self.scene.on_select):
                list = []
                for obj in self.scene.objects:
                    hit = self.scene.objects[obj].select(eye, ray_dir)
                    if hit:
                        list.append(self.scene.objects[obj])
                if list:
                    self.scene.on_select(list)
            if not (self.shift_down or self.ctrl_down):
                self.scene._check_click_plane(eye, ray_dir)

        if event.button == 2:
            self.middle_button_down = True
        if event.button == 3:
            self.right_button_down = True

    def button_release(self, event: Any) -> None:
        """Relese the mouse button."""
        if event.button == 1:
            self.left_button_down = False
        if event.button == 2:
            self.middle_button_down = False
        if event.button == 3:
            self.right_button_down = False

    def scroll(self, event: Any) -> None:
        """Mouse scroll event."""
        self.scene.active_camera.mouse_wheel(event.delta_y)

    def mouse_move(self, event: Any) -> None:
        """Mouse move."""
        camera = self.scene.active_camera
        button = None
        if self.right_button_down:
            button = BUTTON_RIGHT
        if self.middle_button_down:
            button = BUTTON_MIDDLE
        if self.left_button_down:
            button = BUTTON_LEFT
        if self.xrel is None:
            self.xrel = event.x
        if self.yrel is None:
            self.yrel = event.y
        xrel = event.x - self.xrel
        yrel = event.y - self.yrel
        self.xrel = event.x
        self.yrel = event.y
        if not button:
            return

        camera.mouse_move(
            button,
            self.shift_down,
            self.ctrl_down,
            event.x,
            event.y,
            xrel,
            yrel,
            self.scene.window_width,
            self.scene.window_height,
        )

    def keyboard_press(self, event: Any) -> int:  # noqa: C901
        """Keyboard press."""
        keyval = event.keyval
        print(keyval == Gdk.KEY_h)
        if keyval == Gdk.KEY_Escape:
            self.scene.terminate()
            return 2

        if keyval == Gdk.KEY_Shift_L:
            self.shift_down = True

        if keyval == Gdk.KEY_Control_L:
            self.ctrl_down = True

        if keyval == Gdk.KEY_c:
            p = self.scene.active_camera.perspective
            self.scene.active_camera.perspective = not p

        if keyval == Gdk.KEY_space:
            for clock in self.scene.clocks:
                c = self.scene.clocks[clock]
                logging.debug(f"Pause clock [{clock}]")
                c.pause()

        if keyval == Gdk.KEY_h:
            self.scene.huds['_help']._visible = not self.scene.huds['_help'].visible

        if keyval == Gdk.KEY_w:
            for obj in self.scene.objects:
                self.scene.objects[obj].toggle_wireframe()

        if keyval in [Gdk.KEY_F2, Gdk.KEY_F3]:
            active = 0
            for i in range(len(self.scene.cameras)):
                if self.scene.cameras[i].active:
                    active = i
            if keyval == Gdk.KEY_F2:
                active -= 1
            if keyval == Gdk.KEY_F3:
                active += 1
            if active < 0:
                active = len(self.scene.cameras) - 1
            if active > len(self.scene.cameras) - 1:
                active = 0
            self.scene.active_camera = self.scene.cameras[active]
            for i in range(len(self.scene.cameras)):
                self.scene.cameras[i].active = i == active

        return 0



"""
Main Scene Handler.

Scene is the universe. Everything about Payton happens inside a Scene.
"""

# pylama:ignore=C901
import ctypes
import logging
import os
import sys
import time
from dataclasses import dataclass
from typing import Any, Callable, Dict, List, Optional, Tuple, TypeVar

import numpy as np
import sdl2
from OpenGL.GL import (
    GL_CLAMP_TO_EDGE,
    GL_COLOR_BUFFER_BIT,
    GL_DEPTH_ATTACHMENT,
    GL_DEPTH_BUFFER_BIT,
    GL_DEPTH_COMPONENT,
    GL_DEPTH_TEST,
    GL_FLOAT,
    GL_FRAMEBUFFER,
    GL_FRAMEBUFFER_BINDING,
    GL_LESS,
    GL_MAJOR_VERSION,
    GL_MINOR_VERSION,
    GL_NEAREST,
    GL_NONE,
    GL_TEXTURE1,
    GL_TEXTURE_CUBE_MAP,
    GL_TEXTURE_CUBE_MAP_NEGATIVE_X,
    GL_TEXTURE_CUBE_MAP_NEGATIVE_Y,
    GL_TEXTURE_CUBE_MAP_NEGATIVE_Z,
    GL_TEXTURE_CUBE_MAP_POSITIVE_X,
    GL_TEXTURE_CUBE_MAP_POSITIVE_Y,
    GL_TEXTURE_CUBE_MAP_POSITIVE_Z,
    GL_TEXTURE_MAG_FILTER,
    GL_TEXTURE_MIN_FILTER,
    GL_TEXTURE_WRAP_R,
    GL_TEXTURE_WRAP_S,
    GL_TEXTURE_WRAP_T,
    GL_TRIANGLES,
    GL_VERSION,
    glActiveTexture,
    glBindFramebuffer,
    glBindTexture,
    glBindVertexArray,
    glClear,
    glClearColor,
    glDepthFunc,
    glDisable,
    glDrawArrays,
    glDrawBuffer,
    glEnable,
    glFramebufferTexture,
    glGenFramebuffers,
    glGenTextures,
    glGenVertexArrays,
    glGetIntegerv,
    glGetString,
    glReadBuffer,
    glTexImage2D,
    glTexParameteri,
    glViewport,
)

from payton.math.geometry import (
    distance_native,
    raycast_box_intersect,
    raycast_plane_intersect,
    raycast_triangle_intersect,
)
from payton.math.vector import Vector3D
from payton.scene.camera import Camera
from payton.scene.clock import Clock
from payton.scene.collision import CollisionTest
from payton.scene.controller import Controller, GUIController, SceneController
from payton.scene.geometry.base import Object
from payton.scene.grid import Grid
from payton.scene.gui import Hud, Shape2D
from payton.scene.gui.help import help_win
from payton.scene.light import Light
from payton.scene.physics import physics_client
from payton.scene.receiver import Receiver
from payton.scene.shader import (
    DEFAULT_SHADER,
    PARTICLE_SHADER,
    SHADOW_SHADER,
    Shader,
    background_fragment_shader,
    background_vertex_shader,
    default_fragment_shader,
    default_vertex_shader,
    depth_fragment_shader,
    depth_geometry_shader,
    depth_vertex_shader,
    particle_fragment_shader,
    particle_geometry_shader,
    particle_vertex_shader,
)
from payton.scene.types import CPlane

try:
    import pybullet
except ModuleNotFoundError:
    pass

S = TypeVar("S", bound="Scene")
# Shadow Qualities
SHADOW_NONE = 0
SHADOW_LOW = 512
SHADOW_MID = 1024
SHADOW_HIGH = 2048


@dataclass
class PhysicsParams:
    """Global physics parameters for the scene."""

    gravity_x: float = 0
    gravity_y: float = 0
    gravity_z: float = -9.8


class Scene(Receiver):
    """Main Scene where everything happens."""

    def __init__(
        self,
        width: int = 800,
        height: int = 600,
        on_select: Optional[Callable] = None,
        **kwargs: Any,
    ) -> None:
        """Initialize Scene.

        Scene is the playground of the whole show!

        Keyword arguments:
        width -- Window width for the Scene (default 800)
        height -- Window height for the Scene (default 600)
        on_select -- Callback function to handle object selections in the scene

        On Select method definition:

            def select(selected_object_list: List[Object]):
                pass
        """
        self.__fps_counter = 0
        self.fps = 0
        _env_width = os.getenv("SDL_WINDOW_WIDTH", None)
        _env_height = os.getenv("SDL_WINDOW_HEIGHT", None)
        if _env_width:
            width = int(_env_width)
        if _env_height:
            height = int(_env_height)

        # All objects list
        self.objects: Dict[str, Object] = {}
        # All Huds (Heads Up Display)
        self.huds: Dict[str, Hud] = {"_help": Hud(width=width, height=height)}
        self.huds["_help"].add_child("help", help_win())
        self.huds["_help"].hide()

        self.__timer = -1.0
        self.cameras: List[Camera] = []
        self.cameras.append(Camera(active=True, viewport_size=[width, height, 0]))
        self._physics_params = PhysicsParams(0, 0, -10)
        self.clocks: Dict[str, Clock] = {}

        if physics_client is not None:
            pybullet.setGravity(
                self._physics_params.gravity_x,
                self._physics_params.gravity_y,
                self._physics_params.gravity_z,
            )
            pybullet.setPhysicsEngineParameter(
                numSolverIterations=10, minimumSolverIslandSize=1024
            )
            pybullet.setTimeStep(1.0 / 120.0)
            pybullet.connect(pybullet.DIRECT)
            pybullet.configureDebugVisualizer(pybullet.COV_ENABLE_KEYBOARD_SHORTCUTS, 0)
            pybullet.configureDebugVisualizer(pybullet.COV_ENABLE_GUI, 0)
            self.create_clock("_bullet_physics", 1.0 / 120.0, self._step_physics, True)

        self.hudcam = Camera(
            active=True,
            position=[0, 0, 1.0],
            up=[0, 1.0, 0],
            perspective=False,
        )

        self.active_camera = self.cameras[0]

        self.lights: List[Light] = []
        self.lights.append(Light())

        self.grid = Grid()
        self.controller = Controller()
        self.controller.add_controller(GUIController())
        self.controller.add_controller(SceneController())
        self.background = Background()
        self.shaders: Dict[str, Shader] = {
            DEFAULT_SHADER: Shader(
                fragment=default_fragment_shader, vertex=default_vertex_shader
            ),
            PARTICLE_SHADER: Shader(
                fragment=particle_fragment_shader,
                vertex=particle_vertex_shader,
                geometry=particle_geometry_shader,
            ),
            SHADOW_SHADER: Shader(
                fragment=depth_fragment_shader,
                vertex=depth_vertex_shader,
                geometry=depth_geometry_shader,
            ),
        }
        self.shaders["depth"]._depth_shader = True

        # SDL Related Stuff

        self.window = None
        self.window_width = width
        self.window_height = height
        self._context = None
        self._mouse = [0, 0]
        self._shift_down = False
        self._ctrl_down = False
        self._rotate = False
        self.collisions: Dict[str, CollisionTest] = {}
        self._click_planes: List[CPlane] = []

        self.on_select = on_select
        self.depth_map = 0
        self.depth_map_fbo = 0
        # Main running state
        self.running = False
        self._render_lock = False
        self._shadow_quality = SHADOW_MID
        self._shadow_samples = 20

    def _step_physics(self, period: float, total: float) -> None:
        pybullet.stepSimulation()
        for child in self.objects:
            self.objects[child]._bullet_physics()

    @property
    def shadow_samples(self) -> int:
        """Return number of shadow samples."""
        return self._shadow_samples

    @shadow_samples.setter
    def shadow_samples(self, samples: int) -> None:
        """Set shadow samples.

        Keyword arguments:
        samples -- Number of shadow samples (default 20)
        """
        self._shadow_samples = samples % 21
        if self._shadow_samples == 0:
            self._shadow_samples = 1

    @property
    def shadow_quality(self) -> int:
        """Return the shadow quality integer."""
        return self._shadow_quality

    @shadow_quality.setter
    def shadow_quality(self, quality: int) -> None:
        """Set shadow quality.

        Shadow quality is simply the resolution of shadow texture in pixels.
        The higher the resolution gets, the better the shadow is. But that also
        effects the performance in every cycle.

        Keyword arguments:
        quality -- SHADOW_NONE, SHADOW_LOW, SHADOW_MID, SHADOW_HIGH (default SHADOW_MID)
        """
        self._shadow_quality = quality

    def add_click_plane(
        self,
        plane_point: Vector3D,
        plane_normal: Vector3D,
        callback: Callable[[Vector3D], Any],
    ) -> None:
        """Add a click plane to the scene.

        Click planes are infinite 2D planes that can grep and react to clicks.
        You can create a CAD playground or any other stuff that involves selecting
        points in space instead of objects.

        Keyword arguments:
        plane_point -- A point on the surface of the plane.
        plane_normal -- Normal of the plane
        callback -- Callback method to call when a click occurs on the plane

        Callback method definition:

            def hit(hit_point: Vector):
                pass
        """
        self._click_planes.append((plane_point, plane_normal, callback))

    def _check_click_plane(self, eye: Vector3D, vector: Vector3D) -> None:
        """Check if any click planes registered has received a click event."""
        for click_plane in self._click_planes:
            hit = raycast_plane_intersect(eye, vector, click_plane[0], click_plane[1])
            if hit is None:
                continue
            _hit = hit.tolist()
            click_plane[2](_hit[:3])

    def _render_3d_scene(
        self, shadow_round: bool = False, shader: str = DEFAULT_SHADER
    ) -> None:
        """
        Render the 3D Scene.

        Keyword arguments:
        shadow_round -- Is this render for shadow map creation? (default False)
        shader -- Shader to use for the render pass (default DEFAULT_SHADER)
        """
        light_count = len(self.lights)
        lit = light_count > 0
        if not lit:
            return

        proj, view = self.active_camera.render()
        _shader = self.shaders[shader]
        _shader.set_vector3("camera_pos", self.active_camera.position)
        if view is None:
            if not shadow_round:
                _shader.set_int("view_mode", 1)
        else:
            _shader.set_matrix4x4_np("view", view)
            if not shadow_round:
                _shader.set_int("view_mode", 0)

        _shader.set_float("far_plane", self.lights[0].shadow_far_plane)
        _shader.set_matrix4x4_np("projection", proj)
        light_array = [light.position for light in self.lights]
        lcolor_array = [light.color for light in self.lights]
        light_array_np = np.array(light_array, dtype=np.float32)
        lcolor_array_np = np.array(lcolor_array, dtype=np.float32)
        _shader.set_vector3_array_np("light_pos", light_array_np, light_count)

        if not shadow_round:
            _shader.set_int("LIGHT_COUNT", light_count)
            _shader.set_int("samples", self._shadow_samples)
            _shader.set_vector3_array_np("light_color", lcolor_array_np, light_count)
        if self.shadow_quality > 0 and not shadow_round:
            _shader.set_int("shadow_enabled", 1)
        if shadow_round:
            shadow_matrices = self.lights[0].shadow_matrices
            for i, mat in enumerate(shadow_matrices):
                _shader.set_matrix4x4_np("shadowMatrices[{}]".format(i), mat)
        elif self.depth_map > -1:
            glActiveTexture(GL_TEXTURE1)
            glBindTexture(GL_TEXTURE_CUBE_MAP, self.depth_map)
            _shader.set_int("depthMap", 1)

        if not shadow_round and shader == DEFAULT_SHADER:
            self.grid.render(lit, self.shaders[DEFAULT_SHADER])

        for object in self.objects.values():
            if object.shader == shader or (
                shadow_round and object.shader != PARTICLE_SHADER
            ):
                object.render(lit, _shader)

    def _render(self) -> None:
        """
        Render the scene.

        This is the core render method that renders everything to the SDL buffer
        """
        self.shaders["default"].use()
        self._render_lock = True
        glEnable(GL_DEPTH_TEST)
        glDepthFunc(GL_LESS)

        glClearColor(0.1, 0.1, 0.1, 1.0)
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT)

        if self.shadow_quality > 0:
            glViewport(0, 0, self._shadow_quality, self._shadow_quality)
            default_id = glGetIntegerv(GL_FRAMEBUFFER_BINDING)
            glBindFramebuffer(GL_FRAMEBUFFER, self.depth_map_fbo)
            glClear(GL_DEPTH_BUFFER_BIT)
            self.shaders[SHADOW_SHADER].use()
            self._render_3d_scene(True, SHADOW_SHADER)
            glBindFramebuffer(GL_FRAMEBUFFER, default_id)
            self.shaders[SHADOW_SHADER].end()

        # Render background
        glViewport(0, 0, self.window_width, self.window_height)
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT)
        self.background.render()
        glEnable(GL_DEPTH_TEST)
        glDepthFunc(GL_LESS)

        # Render scene
        lit = len(self.lights) > 0
        self.shaders[DEFAULT_SHADER].use()
        self._render_3d_scene(False, DEFAULT_SHADER)

        # Render HUD
        for object in self.huds:
            self.huds[object].render(lit, self.shaders[DEFAULT_SHADER])
        self.shaders[DEFAULT_SHADER].end()

        self.shaders[PARTICLE_SHADER].use()
        self._render_3d_scene(False, PARTICLE_SHADER)
        self.shaders[PARTICLE_SHADER].end()

        self._render_lock = False

        for test in self.collisions.values():
            test.check()

        self.__fps_counter += 1
        if self.__timer == -1:
            self.__timer = time.time()
        diff = time.time() - self.__timer
        if diff > 1:
            self.__timer = time.time()
            self.fps = self.__fps_counter
            self.__fps_counter = 0

    def add_collision_test(self, name: str, tester: CollisionTest) -> None:
        """
        Add a collision tester to the scene.

        Keyword arguments:
        name -- Name of the collision tester to be added
        tester -- CollisionTest object to be added
        """
        if not isinstance(tester, CollisionTest):
            logging.error("tester must be an instance of CollisionTest")
            return

        self.collisions[name] = tester

    def add_object(self, name: str, obj: Object) -> bool:
        """
        Add an object to the scene.

        Object can be a HUD object, 2D Shape or 3D Mesh.

        Keyword arguments:
        name -- Name of the object to add to the scene
        obj -- 3D Object to be added.
        """
        if not isinstance(obj, Object):
            logging.error("Given object is not an instance of `scene.Object`")
            return False

        if name in self.objects:
            logging.error(f"Given object name [{name}] already exists")
            return False

        if isinstance(obj, Shape2D):
            logging.error("2D Shapes can't be added directly to the scene")
            return False

        if isinstance(obj, Hud):
            """Huds must be rendered in a different loop after rendering
            all objects"""
            if self._render_lock:
                while self._render_lock:
                    # Wait for render loop to release the lock
                    # TODO: This can cause a possible deadlock.
                    continue
            self.huds[name] = obj
            obj.name = name
            obj.set_size(self.window_width, self.window_height)
            return True

        if self._render_lock:
            while self._render_lock:
                # Wait for render loop to release the lock
                # TODO: This can cause a possible deadlock.
                continue
        self.objects[name] = obj
        obj.name = name
        return True

    def add_camera(self, camera: Camera) -> bool:
        """Add a camera to the scene and return True if successful.

        If an invalid type of object is given to the scene as an camera,
        it does not break the scene as the scene already has a working camera
        to operate. Instead, failure returns False

        Keyword arguments:
        camera -- Camera object to be added
        """
        if not isinstance(camera, Camera):
            logging.error("Camera is not an instance of `scene.camera`")
            return False

        self.cameras.append(camera)
        return True

    def create_camera(self) -> None:
        """Create a dummy camera with defaults."""
        self.cameras.append(
            Camera(viewport_size=[self.window_width, self.window_height, 0])
        )

    def create_clock(
        self,
        name: str,
        period: float,
        callback: Callable[[float, float], None],
        non_stop: bool = False,
    ) -> None:
        """Create a Clock in the scene.

        Clocks are asyn calls that are triggered in given periods.
        Clocks should be written effectively to take as less time as possible.
        If a clock can not be completed within the given period, there can be two parallel
        processing interfering with eachother.

        Keyword arguments:
        name -- Name of the clock to create
        period -- Period of the callback function in seconds
        callback -- Callback method to call for each "tick" of the clock.

        Callback method definition:

            def logger(period_of_the_clock: float, total_seconds_passed: float)
                pass
        """
        if name in self.clocks:
            logging.error(f"A clock named {name} already exists")
            return

        c = Clock(period, callback, non_stop)
        self.clocks[name] = c

    def _clear_context(self) -> None:
        for shader in self.shaders.values():
            shader.program = -1
        self.background._shader.program = -1
        self.background._vao = None
        self.grid.destroy()
        for hud in self.huds.values():
            hud.destroy()
        # As Payton clocks are basically threads,
        # best practice is to create new threads as Python does not let
        # a thread to be started twice by design.
        new_clocks = {}
        for clock_name in self.clocks:
            clock = self.clocks[clock_name]
            new_clock = Clock(clock.period, clock.callback)
            new_clock._total_time = clock._total_time
            new_clocks[clock_name] = new_clock

        self.clocks = new_clocks

    def to_dict(self) -> Dict[str, Any]:
        """Convert everything inside the Scene to Dictionary.

        This method is better to be called outside the render cycle as it can be
        very costly.
        """
        return {
            "objects": {name: self.objects[name].to_dict() for name in self.objects},
            "lights": [light.to_dict() for light in self.lights],
            "cameras": [camera.to_dict() for camera in self.cameras],
        }

    def raycast_intersect(
        self,
        start: Vector3D,
        direction: Vector3D,
        box_only: bool = True,
        exempt_objects: Optional[List[Object]] = None,
    ) -> Optional[Tuple[Object, Vector3D]]:
        """Raycast a vector to the scene.

        Assume that there is a vector in space starting from "start"
        and going in "vector" direction. This method checks through the
        scene objects and returns the list of objects the vector hits
        with their distances.

        Keyword arguments:
        start -- Start of the vector
        direction -- Direction of the vector
        box_only -- Only make an axis-aligned bounding box collusion check (default True)
        exempt_objects -- Exempt these objects from being tested
        """
        shortest = [0.0, 0.0, 0.0]
        dist_best = -1.0
        hit_obj = None
        for obj in self.objects.values():
            if exempt_objects is not None and obj in exempt_objects:
                continue
            box_hit = raycast_box_intersect(
                start, direction, obj.bounding_box[0], obj.bounding_box[1]
            )
            if box_hit is None:
                continue
            if box_only:
                dist = distance_native(box_hit, start)
                if dist_best == -1.0 or dist < dist_best:
                    dist_best = dist
                    shortest = box_hit
                    hit_obj = obj
                    continue
            for p1, p2, p3 in zip(*[iter(obj.absolute_vertices())] * 3):
                ip_np, _ = raycast_triangle_intersect(start, direction, p1, p2, p3)
                if ip_np is None:
                    continue
                ip = ip_np.tolist()
                dist = distance_native(ip, start)
                if dist_best == -1.0 or dist < dist_best:
                    dist_best = dist
                    shortest = ip
                    hit_obj = obj
        if hit_obj is None:
            return None
        return hit_obj, shortest

    def _init_runtime(self, start_clocks: bool = False) -> bool:
        version = glGetString(GL_VERSION).decode("utf-8")
        ogl_major = glGetIntegerv(GL_MAJOR_VERSION)
        ogl_minor = glGetIntegerv(GL_MINOR_VERSION)
        print(f"gl-int: {ogl_major}.{ogl_minor} - {version}")

        v = (ogl_major * 10) + ogl_minor
        if v < 33:  # OpenGL 3.3
            print(
                """
Sorry, Payton couldn't create an OpenGL 3.3 Context
This can be related to an old graphics card / an old driver issue.

Payton requires at least OpenGL 3.3 support and above."""
            )
            print(f"OpenGL Version installed: {ogl_major}.{ogl_minor}")
            sys.exit(1)

        # Fix aspect ratios of cameras
        for camera in self.cameras:
            camera.aspect_ratio = self.window_width / self.window_height * 1.0

        for hud in self.huds.values():
            hud.set_size(self.window_width, self.window_height)

        for clock in self.clocks:
            self.clocks[clock].start()
            if start_clocks:
                self.clocks[clock].pause()

        # if shadows
        # @TODO Creating the shadow cubemap here is not a good idea
        self.depth_map_fbo = glGenFramebuffers(1)
        self.depth_map = glGenTextures(1)
        side_map = [
            GL_TEXTURE_CUBE_MAP_POSITIVE_X,
            GL_TEXTURE_CUBE_MAP_NEGATIVE_X,
            GL_TEXTURE_CUBE_MAP_POSITIVE_Y,
            GL_TEXTURE_CUBE_MAP_NEGATIVE_Y,
            GL_TEXTURE_CUBE_MAP_POSITIVE_Z,
            GL_TEXTURE_CUBE_MAP_NEGATIVE_Z,
        ]
        glBindTexture(GL_TEXTURE_CUBE_MAP, self.depth_map)
        for i in side_map:
            glTexImage2D(
                i,
                0,
                GL_DEPTH_COMPONENT,
                self.shadow_quality,
                self.shadow_quality,
                0,
                GL_DEPTH_COMPONENT,
                GL_FLOAT,
                None,
            )

        glTexParameteri(GL_TEXTURE_CUBE_MAP, GL_TEXTURE_MAG_FILTER, GL_NEAREST)
        glTexParameteri(GL_TEXTURE_CUBE_MAP, GL_TEXTURE_MIN_FILTER, GL_NEAREST)
        glTexParameteri(GL_TEXTURE_CUBE_MAP, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE)
        glTexParameteri(GL_TEXTURE_CUBE_MAP, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE)
        glTexParameteri(GL_TEXTURE_CUBE_MAP, GL_TEXTURE_WRAP_R, GL_CLAMP_TO_EDGE)

        default_id = glGetIntegerv(GL_FRAMEBUFFER_BINDING)
        glBindFramebuffer(GL_FRAMEBUFFER, self.depth_map_fbo)
        glFramebufferTexture(
            GL_FRAMEBUFFER,
            GL_DEPTH_ATTACHMENT,
            self.depth_map,
            0,
        )
        glDrawBuffer(GL_NONE)
        glReadBuffer(GL_NONE)
        glBindFramebuffer(GL_FRAMEBUFFER, default_id)
        for shader in self.shaders.values():
            shader.build()

        return True

    def run(self, start_clocks: bool = False) -> int:
        """Run the show."""
        if sdl2.SDL_Init(sdl2.SDL_INIT_VIDEO) != 0:
            return -1

        sdl2.SDL_GL_SetAttribute(
            sdl2.SDL_GL_CONTEXT_PROFILE_MASK, sdl2.SDL_GL_CONTEXT_PROFILE_CORE
        )
        multisample_buffers = os.getenv("GL_MULTISAMPLEBUFFERS", None)
        multisample_samples = os.getenv("GL_MULTISAMPLESAMPLES", None)
        if multisample_buffers:
            sdl2.SDL_GL_SetAttribute(
                sdl2.SDL_GL_MULTISAMPLEBUFFERS, int(multisample_buffers)
            )
        if multisample_samples:
            sdl2.SDL_GL_SetAttribute(
                sdl2.SDL_GL_MULTISAMPLESAMPLES, int(multisample_samples)
            )

        self.window = sdl2.SDL_CreateWindow(
            b"Payton Scene",
            sdl2.SDL_WINDOWPOS_UNDEFINED,
            sdl2.SDL_WINDOWPOS_UNDEFINED,
            int(self.window_width),
            int(self.window_height),
            sdl2.SDL_WINDOW_OPENGL | sdl2.SDL_WINDOW_RESIZABLE,
        )

        if not self.window:
            return -1

        self._context = sdl2.SDL_GL_CreateContext(self.window)
        sdl2.SDL_GL_SetSwapInterval(0)
        self.event = sdl2.SDL_Event()
        self.running = True

        self._init_runtime()
        while self.running:
            while sdl2.SDL_PollEvent(ctypes.byref(self.event)) != 0:
                if (
                    self.event.type == sdl2.SDL_WINDOWEVENT
                    and self.event.window.event == sdl2.SDL_WINDOWEVENT_RESIZED
                ):
                    self.window_width = self.event.window.data1
                    self.window_height = self.event.window.data2
                    glViewport(0, 0, self.window_width, self.window_height)

                    for ob in self.cameras:
                        ob.aspect_ratio = self.window_width / self.window_height
                        ob._viewport_size = [self.window_width, self.window_height, 0]
                    for hud in self.huds:
                        self.huds[hud].set_size(self.window_width, self.window_height)
                self.controller.keyboard(self.event, self)
                self.controller.mouse(self.event, self)

            self._render()

            sdl2.SDL_GL_SwapWindow(self.window)
            sdl2.SDL_Delay(1)

        for obj in self.objects:
            self.objects[obj].destroy()

        for clock in self.clocks:
            self.clocks[clock].kill()
            self.clocks[clock]._hold = True

        sdl2.SDL_GL_DeleteContext(self._context)
        sdl2.SDL_DestroyWindow(self.window)
        self.window = None
        sdl2.SDL_Quit()
        self._clear_context()
        return 0

    def terminate(self) -> None:
        """Terminate scene."""
        self.running = False
        for clock in self.clocks:
            logging.debug(f"Kill clock [{clock}]")
            self.clocks[clock].kill()
            self.clocks[clock].join()


class Background:
    """(Shader code and idea derived from the original work of.

    https://www.cs.princeton.edu/~mhalber/blog/ogl_gradient/)
    """

    def __init__(
        self,
        top_color: Optional[Vector3D] = None,
        bottom_color: Optional[Vector3D] = None,
        **kwargs: Dict[str, Any],
    ):
        """Initialize the background.

        Keyword arguments:
        top_color -- Gradient color at the top of the viewport
        bottom_color -- Gradient color at the bottom of the viewport
        """
        self.top_color = [0.0, 0.0, 0.0, 1.0] if top_color is None else top_color
        self.bottom_color = (
            [0.0, 0.1, 0.2, 1.0] if bottom_color is None else bottom_color
        )
        variables = ["top_color", "bot_color"]
        self._shader = Shader(
            fragment=background_fragment_shader,
            vertex=background_vertex_shader,
            variables=variables,
        )
        self._vao = None
        self.visible = True

    def set_time(self, hour: int, minute: int) -> None:
        """Background can mimic a background color based on the time of the date.

        It is not accurate but at least gives a small impression

        Keyword arguments:
        hour -- Hour as integer (24 hour format)
        minute -- Minute as integer
        """
        hour %= 24
        minute %= 60
        color_scheme = {
            (0, 3): [
                [24 / 255, 24 / 255, 48 / 255, 1.0],
                [24 / 255, 48 / 255, 72 / 255, 1.0],
            ],
            (3, 9): [
                [24 / 255, 48 / 255, 72 / 255, 1.0],
                [96 / 255, 168 / 255, 192 / 255, 1.0],
            ],
            (9, 15): [
                [96 / 255, 168 / 255, 192 / 255, 1.0],
                [144 / 255, 192 / 255, 240 / 255, 1.0],
            ],
            (15, 21): [
                [144 / 255, 192 / 255, 240 / 255, 1.0],
                [48 / 255, 72 / 255, 120 / 255, 1.0],
            ],
            (21, 24): [
                [48 / 255, 72 / 255, 120 / 255, 1.0],
                [24 / 255, 24 / 255, 48 / 255, 1.0],
            ],
        }
        for color in color_scheme:
            if hour >= color[0] and hour < color[1]:
                hours = color[1] - color[0]
                minutes = hours * 60
                minute = ((hour - color[0]) * 60) + minute
                xdist = (
                    (color_scheme[color][1][0] - color_scheme[color][0][0]) / minutes
                ) * minute
                ydist = (
                    (color_scheme[color][1][1] - color_scheme[color][0][1]) / minutes
                ) * minute
                zdist = (
                    (color_scheme[color][1][2] - color_scheme[color][0][2]) / minutes
                ) * minute

                self.bottom_color = color_scheme[color][0]
                self.top_color = [
                    color_scheme[color][0][0] + xdist,
                    color_scheme[color][0][1] + ydist,
                    color_scheme[color][0][2] + zdist,
                    1.0,
                ]
                return

    def render(self) -> None:
        """Render the background."""
        if not self.visible:
            return

        if not self._vao:
            self._vao = glGenVertexArrays(1)
            glBindVertexArray(self._vao)
            self._shader.build()
            glBindVertexArray(0)

        glDisable(GL_DEPTH_TEST)

        self._shader.use()
        self._shader.set_vector4_np(
            "top_color", np.array(self.top_color, dtype=np.float32)
        )
        self._shader.set_vector4_np(
            "bot_color", np.array(self.bottom_color, dtype=np.float32)
        )
        glBindVertexArray(self._vao)
        glDrawArrays(GL_TRIANGLES, 0, 3)
        glBindVertexArray(0)
        self._shader.end()
        glEnable(GL_DEPTH_TEST)



"""Physics resolver."""

physics_client = None

try:
    import pybullet

    print("Bullet Physics enabled")
    physics_client = pybullet.connect(pybullet.DIRECT)
except ModuleNotFoundError:
    print("Bullet Physics is not installed, continue without Physics support")


class PhysicsException(Exception):
    """Raise this exception when needed."""

    pass



import math
import os
from textwrap import wrap
from typing import Any, Callable, Dict, List, Optional, Tuple, TypeVar, cast

import numpy as np
from OpenGL.GL import GL_DEPTH_TEST, glDisable, glEnable
from PIL import Image, ImageDraw
from PIL.ImageFont import ImageFont, truetype

from payton.math.functions import ortho
from payton.math.vector import Vector3D
from payton.scene.geometry.base import Object
from payton.scene.geometry.mesh import Mesh
from payton.scene.shader import Shader

S2 = TypeVar("S2", bound="Shape2D")


def text_size(s: str, font: ImageFont | None) -> Tuple[int, int]:
    timg = Image.new("RGBA", (1, 1))
    d = ImageDraw.Draw(timg)
    size = d.textbbox((0, 0), s, font=font)
    x = size[2] - size[0]
    y = size[3] - size[1]
    return int(x), int(y)


class Shape2D(Mesh):
    def __init__(
        self,
        position: List[int],
        size: List[int],
        on_click: Optional[Callable] = None,
        opacity: float = 0.1,
        **kwargs: Any,
    ):
        """Initialize 2D Shape

        Keyword arguments:
        position -- Position of the shape on the screen (x, y, and optionally Z as integer)
        size -- Size of the shape (w, h as integer)
        on_click -- On Click event callback (takes no arguments)
        opacity -- Opacity of the shape
        """
        super().__init__(**kwargs)
        self.material.opacity = opacity
        position = list(position)
        if len(position) == 2:
            position.append(0)
        position[2] = -1
        self.__position = position
        self.position = [float(x) for x in self.__position]
        self.size = size
        self.on_click: Optional[Callable] = on_click
        self._font: ImageFont | None = None
        self.parent: Any = None
        self._scene_width: int = 0
        self._scene_height: int = 0

    @property
    def physics(self) -> bool:
        return False

    def add_child(self, name: str, obj: Object) -> bool:
        """Add a child shape to this shape object.

        This method overrides the parent by arguments. Child object's position
        follows the parent.

        Keyword arguments:
        name -- Name of the object to be added
        obj -- Shape2D Object to add.
        """
        res = super().add_child(name, cast(Object, obj))
        if not res:
            return res
        if isinstance(obj, Shape2D):
            obj.parent = self
        return res

    def draw(self) -> None:
        """Placeholder for draw function"""
        ...

    def draw_text(self) -> None:
        """Placeholder for draw text function"""
        ...

    def render(
        self,
        lit: bool,
        shader: Shader,
        parent_matrix: Optional[np.ndarray] = None,
        _primitive: Optional[int] = None,
    ) -> None:
        """Render cycle for the Shape 2D

        We override the `lit` parameter for the super.
        This is not actually intended to be used publicly except if you want to
        setup your own render cycle.
        """
        super().render(False, shader, parent_matrix)

    @property
    def font(self) -> None | ImageFont:
        """Return the fond of the shape"""
        if self._font is not None:
            return self._font
        if self.parent is not None:
            return self.parent.font
        return None

    @font.setter
    def font(self, font: ImageFont) -> None:
        """Set the font of the Shape

        Keyword arguments:
        font -- ImageFont object
        """
        self._font = font
        self.draw_text()

    def _set_parent_size(self, w: int, h: int) -> None:
        self._parent_width = w
        self._parent_height = h
        self._init = False
        self.draw()

        for child in self.children.values():
            cast("Shape2D", child)._set_parent_size(self.size[0], self.size[1])

    def click(self, x: int, y: int) -> Optional[Mesh]:
        """Check for click event

        Keyword arguments:
        x -- Cursor X
        y -- Cursor Y
        """
        if not callable(self.on_click):
            for child in self.children:
                c = cast(Mesh, self.children[child]).click(x, y)
                if c:
                    return cast(Mesh, self.children[child])
            return None

        if self._model_matrix is None:
            return None
        mm = self._model_matrix[3]
        if x > mm[0] and x < mm[0] + self.size[0] and y > mm[1] and y < mm[1] + self.size[1]:
            self.on_click()
            return self
        return None


class Rectangle(Shape2D):
    def __init__(
        self,
        position: List[int],
        size: List[int],
        **kwargs: Any,
    ):
        """Initialize Rectangle

        Keyword arguments:
        position -- Tuple as [X, Y, Z] positions
        size -- Size of the rectangle
        """
        super().__init__(position=position, size=size, **kwargs)
        self._init: bool = False
        self.draw()

    def draw(self) -> None:
        """Create the rectangle polygons"""
        w, h = self.size
        self.clear_triangles()
        if not self._init:
            self.add_triangle(
                [[0, 0, 1], [w, h, 1], [w, 0, 1]],
                texcoords=[[0, 0], [1, 1], [1, 0]],
            )
            self.add_triangle(
                [[0, 0, 1], [0, h, 1], [w, h, 1]],
                texcoords=[[0, 0], [0, 1], [1, 1]],
            )
            self._init = True


class Text(Rectangle):
    def __init__(
        self,
        position: List[int],
        size: List[int],
        label: str = "lorem",
        bgcolor: Optional[Vector3D] = None,
        color: Optional[Vector3D] = None,
        **kwargs: Any,
    ) -> None:
        """Initialize Text object (label)

        Keyword arguments:
        position -- Tuple as [X, Y, Z] positions on the screen
        size -- Size of the text area
        label -- Text / Label to be printed
        bgcolor -- Background color of the text
        color -- Color of the text"""
        super().__init__(position=position, size=size, **kwargs)
        self.__label: str = label
        self.bgcolor: Vector3D = [0, 0, 0, 0]
        if bgcolor is not None:
            self.bgcolor = bgcolor

        self.color: Vector3D = [0, 0, 0]
        if color is not None:
            self.color = color

        self.crop = [0, 0, 0, 0]
        self._init_text: bool = False
        self._max_char_width = 0.0

    @property
    def label(self) -> str:
        """Get the label of the text"""
        return self.__label

    @label.setter
    def label(self, label: str) -> None:
        """Set the label of the text

        Keyword arguments:
        label -- Label string
        """
        self.__label = label
        self._init_text = False

    def render(
        self,
        lit: bool,
        shader: Shader,
        parent_matrix: Optional[np.ndarray] = None,
        _primitive: Optional[int] = None,
    ) -> None:
        """Render the text object by initializing the text material if not initialized yet

        Keyword arguments:
        lit -- Is the object illuminated?
        shader -- Shader to use while rendering the object
        parent_matrix -- Parent object's matrix if this is a child object
        _primitive -- override the scene wide primitive (rendering) mode. (Point, Wire, Solid)
        """
        if not self._init_text:
            self.draw_text()
        super().render(lit, shader, parent_matrix)

    @property
    def text_size(self) -> Tuple[int, int]:
        """Return the text size in pixels"""
        if self.font is None:
            return (0, 0)
        x, y = text_size(self.label, self.font)
        return max(int(self.size[0]), x), max(int(self.size[1]), y)

    def wrap(self, width_in_pixels: int) -> None:
        """Word-wrap the text to fit into the given pixel size

        Keyword arguments:
        width_in_pixels -- Width size in pixels to fit the text into
        """
        original = self.label
        self.label = ""
        if self.font is None:
            return
        if self._max_char_width == 0.0:
            self._max_char_width = text_size('_', self.font)[0]
        split_length = int(math.floor(width_in_pixels / self._max_char_width)) - 1
        parts = wrap(original, split_length)
        self.__label = "\n".join(parts)

        self._init_text = False

    def draw_text(self) -> None:
        """Create the text material and initialize the font if needed"""
        if self._init_text:
            return
        bgcolor = (
            int(self.bgcolor[0] * 255),
            int(self.bgcolor[1] * 255),
            int(self.bgcolor[2] * 255),
            int(self.bgcolor[3] * 255),
        )

        img = Image.new("RGBA", self.text_size, color=bgcolor)
        d = ImageDraw.Draw(img, "RGBA")
        color = (int(self.color[0] * 255), int(self.color[1] * 255), int(self.color[2] * 255), 255)
        if self.font is not None:
            d.text((1, 1), self.label, fill=color, font=self.font)
        else:
            d.text((1, 1), self.label, fill=color)

        if any(self.crop):
            img = img.crop((self.crop[0], self.crop[1], self.crop[2], self.crop[3]))

        del d

        self.material._image = img
        self.material.opacity = 1.0
        self.material.refresh()
        self._init_text = True


class Hud(Object):
    def __init__(
        self,
        width: int = 800,
        height: int = 600,
        font: str = "",
        font_size: int = 15,
        **kwargs: Any,
    ) -> None:
        """Initialize HUD Object to render HUD elements.

        You need a HUD in a scene to render 2D shapes on top of your scene

        Keyword arguments:
        width -- Width of the HUD area
        height -- Height of the HUD area
        font -- Font name to use (default: monofonto provided by the package)
        font_size -- Font size in pixels
        """
        super().__init__(**kwargs)
        self.width: int = width
        self.height: int = height
        self._fontname: str = font
        self.children: Dict[str, Object] = {}
        self._font_size: int = font_size
        if self._fontname != "":
            self.set_font(self._fontname, self._font_size)
        self._font: ImageFont | None = None
        self._projection_matrix: Optional[np.ndarray] = None
        try:
            self.set_font(os.path.join(os.path.dirname(os.path.abspath(__file__)), "monofonto.ttf"))
        except OSError:
            # font not found but pillow has better than nothinf font
            pass

    @property
    def font(self) -> ImageFont | None:
        """Return the HUD Image Font"""
        return self._font

    def add_child(self, name: str, obj: Object) -> bool:
        """Add 2D Shape into Hud.

        Note: this is a type ignore due to it's mismatch with Object add_child method

        Keyword arguments:
        name -- Name of the object
        obj -- Shape2D object to add"""
        res = super().add_child(name, obj)
        if isinstance(obj, Shape2D):
            obj._scene_height = self.height
            obj._scene_width = self.width
            obj.parent = self

        if not res:
            return res
        return res

    def render(self, lit: bool, shader: Shader, *_args: Any) -> None:
        """Main render cycle for HUD

        Disables the DEPTH test and draws the objects from parent to child
        """
        if not self.visible:
            return
        lit = False
        glDisable(GL_DEPTH_TEST)
        if self._projection_matrix is None:
            self._projection_matrix = ortho(0, self.width, self.height, 0)

        shader.set_int("view_mode", 1)
        shader.set_matrix4x4_np("projection", self._projection_matrix)
        for child in self.children:
            self.children[child].render(lit, shader, None)
        glEnable(GL_DEPTH_TEST)

    def set_size(self, w: int, h: int) -> None:
        """Set HUD size

        Keyword arguments:
        w -- Width of the HUD
        h -- Height of the HUD
        """
        self.width = w
        self.height = h
        for child in self.children.values():
            if isinstance(child, Shape2D):
                child._set_parent_size(w, h)
                child._init = False
                child.draw()

        self._projection_matrix = None

    def set_font(self, font_name: str, font_size: int = 16) -> None:
        """Set True Type font for the Hud

        Keyword arguments:
        font_name -- Name of the True Type font installed in the system
        font_size -- Size of the font in pixels
        """
        self._font = truetype(font_name, font_size)  # type: ignore



"""Plane geometry module."""

from itertools import product
from typing import Any, List

from payton.scene.geometry.mesh import Mesh
from payton.scene.material import SOLID, WHITE

_BULLET = False
try:
    import pybullet

    _BULLET = True
except ModuleNotFoundError:
    _BULLET = False


class Plane(Mesh):
    """Plane object."""

    def __init__(self, width: float = 1.0, height: float = 1.0, **kwargs: Any) -> None:
        """Initialize a plane surface.

        Keyword arguments:
        width -- Width of the plane
        height -- Height of the plane
        """
        super().__init__(**kwargs)
        width *= 0.5
        height *= 0.5
        self._vertices = [
            [-width, -height, 0],
            [width, -height, 0],
            [width, height, 0],
            [-width, height, 0],
        ]
        self._normals = [[0, 0, 1], [0, 0, 1], [0, 0, 1], [0, 0, 1]]
        self._texcoords = [[-1, -1], [1, -1], [1, 1], [-1, 1]]
        self._indices = [[0, 1, 2], [2, 3, 0]]
        self.material._indices = self._indices

    def _create_collision_shape(self) -> None:
        self._bullet_shape_id = pybullet.createCollisionShape(pybullet.GEOM_PLANE, planeNormal=self.matrix[2][:3])


class MatrixPlane(Mesh):
    """Matrix plane."""

    def __init__(
        self,
        width: float = 1.0,
        height: float = 1.0,
        x: int = 2,
        y: int = 2,
        **kwargs: Any,
    ) -> None:
        """Define a Matrix Plane.

        Matrix plane is an MxN grid matrix where you can change the height / color
        of each individual vertex in the matrix by ease

        Keyword arguments:
        width -- Width of the plane
        height -- Height of the plane
        x -- Number of points in X direction
        y -- Number of points in Y direction
        """
        super().__init__(**kwargs)
        self.width = width
        self.height = height
        self.x = x
        self.y = y
        self.x = max(self.x, 2)
        self.y = max(self.y, 2)
        self.grid: List[List[float]] = []
        # List of List for X, Y and List[float] for color.
        self.color_grid: List[List[List[float]]] = []
        self.populate_grid()

    def update_grid(self) -> None:
        """Update the grid for changes."""
        # TODO This method needs some optimization
        for i, j in product(range(self.x), range(self.y)):
            self._vertices[(self.x * i) + j][2] = self.grid[i][j]
            self._vertex_colors[(self.x * i) + j] = self.color_grid[i][j]
        if self.material.display == SOLID:
            self.fix_normals()
        self.refresh()

    def populate_grid(self) -> None:
        """Turn grid data into plane."""
        self.clear_triangles()

        self.grid = [[0.0] * self.y for _i in range(self.x)]
        self.color_grid = [[WHITE] * self.y for _i in range(self.x)]

        step_x = self.width / (self.x - 1)
        step_y = self.height / (self.y - 1)
        step_u = 1.0 / (self.x - 1)
        step_v = 1.0 / (self.y - 1)

        c_x = self.width / 2.0
        c_y = self.height / 2.0
        for i, j in product(range(self.x), range(self.y)):
            self._vertices.append([(i * step_x) - c_x, (j * step_y) - c_y, self.grid[i][j]])
            self._texcoords.append([i * step_u, j * step_v])
            self._vertex_colors.append(WHITE)

            if i < self.x - 1 and j < self.y - 1:
                left = (self.x * i) + j
                right = (self.x * i) + (j + 1)
                top_left = (self.x * (i + 1)) + j
                top_right = (self.x * (i + 1)) + (j + 1)
                self._indices.append([top_right, right, left])
                self._indices.append([left, top_left, top_right])

        self.material._indices = self._indices
        self.fix_normals()
        self.refresh()



from typing import Any

try:
    import gi

    gi.require_version('Gtk', '3.0')
    from gi.repository import Gdk, Gtk

    _GTK_SUPPORTED = True
except ModuleNotFoundError:
    _GTK_SUPPORTED = False

from payton.scene import Scene
from payton.scene.gtkcontroller import GTKController

if _GTK_SUPPORTED:  # noqa: C901

    class PaytonGTKSceneArea(Gtk.EventBox):
        """Instance of GTk GLArea."""

        def __init__(self, scene: Scene):
            """Initialize the GLArea with the Payton Scene."""
            Gtk.EventBox.__init__(self)
            self._gl_area = Gtk.GLArea()

            self._gl_area.connect("realize", self.on_realize)
            self._gl_area.connect("render", self.render)
            self._gl_area.connect("resize", self.resize)

            self.set_events(Gdk.EventMask.ALL_EVENTS_MASK)

            self.connect("button-press-event", self.button_press)
            self.connect("button-release-event", self.button_release)
            self.connect("motion-notify-event", self.motion_notify)
            self.connect("scroll-event", self.scroll)

            self._gl_area.set_has_depth_buffer(True)
            self._gl_area.set_has_stencil_buffer(True)
            self.add(self._gl_area)
            self.scene = scene
            self.controller = GTKController(self.scene)

        def render(self, area: Gtk.GLArea, ctx: Any) -> bool:
            """Render the scene."""
            ctx.make_current()
            area.make_current()
            self.scene._render()
            area.queue_draw()
            return True

        def scroll(self, widget: "PaytonGTKSceneArea", event: Gdk.EventScroll) -> None:
            """Scroll mouse."""
            self.controller.scroll(event)

        def resize(self, area: Gtk.GLArea, width: int, height: int) -> None:
            """Resize the widget."""
            self.scene.window_width = width
            self.scene.window_height = height
            for ob in self.scene.cameras:
                ob.aspect_ratio = width / height
                ob._viewport_size = [width, height, 0]
            for hud in self.scene.huds:
                self.scene.huds[hud].set_size(width, height)

        def on_realize(self, area: Gtk.GLArea) -> None:
            """Realize (initialize) the area."""
            area.make_current()
            self.scene.window_width = area.get_allocated_width()
            self.scene.window_height = area.get_allocated_height()

            self.scene.running = True
            self.scene._init_runtime()
            self.scene._context = area
            err = self._gl_area.get_error()
            if err:
                print("Something went wrong with the GTK:")
                print(err)
            else:
                print("Widget initialized.")

        def key_press(self, widget: Gtk.Window, event: Gdk.EventKey) -> None:
            """Handle key press."""
            self.controller.keyboard_press(event)

        def key_release(self, widget: Gtk.Window, event: Gdk.EventKey) -> None:
            """Handle key release."""
            self.controller.keyboard_release(event)

        def button_press(self, widget: Gtk.Window, event: Gdk.EventButton) -> None:
            """Handle button press."""
            self.controller.button_press(event)

        def button_release(self, widget: Gtk.Window, event: Gdk.EventButton) -> None:
            """Handle button release event."""
            self.controller.button_release(event)

        def motion_notify(self, widget: Gtk.Window, event: Gdk.EventMotion) -> None:
            """Handle mouse motion."""
            self.controller.mouse_move(event)

else:

    class PaytonGTKSceneArea:  # type: ignore
        """Payton GTK Scene for invalid GTK Imports."""

        def __init__(self, scene: Scene):
            """This shall fail."""
            raise BaseException("PyGTK can not be initialized")



# pylama:ignore=C901
# Wavefront Object File Support
# Only support ascii obj files and without material support.
# So pretty limited.
import logging
import os
from shutil import copyfile
from typing import Any, List

from payton.scene.geometry.mesh import Mesh
from payton.scene.material import DEFAULT, Material
from payton.scene.types import IList, VList


class Wavefront(Mesh):
    def __init__(self, filename: str = "", **kwargs: Any) -> None:
        """Initialize the Wavefront mesh

        Keyword arguments:
        filename -- Filename to the wavefront object file
        """
        super().__init__()
        self.filename: str = filename
        self.path: str = ""
        if self.filename != "":
            self.load_file(self.filename)

    def load_file(self, filename: str) -> bool:
        """Load file into mesh

        Keyword arguments:
        filename -- filename to load"""
        if not os.path.isfile(filename):
            logging.exception(f"File not found {filename}")
            return False

        self.filename = filename
        self.path = os.path.dirname(os.path.abspath(self.filename))
        with open(filename) as f:
            data = f.read()
        self.load(data)
        return True

    @property
    def physics(self) -> bool:
        """Is a physics responding object."""
        return True

    def load_material(self, material_string: str) -> None:
        """Load material from given material definition string

        Please note that, material capabilities of Payton are limited
        and not all material properties are supported.

        Keyword arguments:
        material_string -- Wavefront Material Syntax
        """
        lines = material_string.splitlines()
        material = Material()
        material_name = DEFAULT
        for line in lines:
            parts = line.split(" ")

            if parts[0] == "newmtl":
                if material_name != DEFAULT:
                    self.add_material(material_name, material)
                material = Material()
                material_name = parts[1]

            if parts[0] == "Kd":
                # Currently we only support diffuse color
                material.color = [
                    float(parts[1]),
                    float(parts[2]),
                    float(parts[3]),
                ]

            if parts[0] == "map_Kd":
                rest = " ".join(parts[1:])
                if rest.startswith("/"):
                    if not os.path.exists(rest):
                        base_name = os.path.basename(rest)
                        rest = os.path.join(self.path, base_name)
                    material.texture = rest
                else:
                    material.texture = os.path.join(self.path, rest)

        self.add_material(material_name, material)

    def load_material_file(self, filename: str) -> None:
        """Load wavefront material file (.mtl)

        Keyword arguments:
        filename -- Filename to load
        """
        if not os.path.isfile(filename):
            return
        data = open(filename).read()
        self.load_material(data)

    def load(self, obj_string: str) -> None:
        """Load wavefront object data

        Keyword arguments:
        obj_string -- Wavefront object definitions
        """
        _vertices: VList = []
        _indices: List[IList] = []
        _indice_materials: List[str] = []
        _normals: VList = []
        _texcoords: VList = []
        lines: List[str] = obj_string.splitlines()
        material = DEFAULT
        for line in lines:
            line = line.replace("  ", " ")
            parts = line.split(" ")
            command = parts[0].lower()
            if parts[0] == "mtllib":
                self.load_material_file(f"{self.path}/{parts[1]}")
            if command == "v":
                x = float(parts[1])
                y = float(parts[2])
                z = float(parts[3])
                _vertices.append([x, y, z])
            if command == "vt":
                u = float(parts[1])
                w = 1.0 - float(parts[2]) if len(parts) > 2 else 0
                _texcoords.append([u, w])
            if command == "vn":
                x = float(parts[1])
                y = float(parts[2])
                z = float(parts[3])
                _normals.append([x, y, z])
            if command == "usemtl" and parts[1] in self.materials:
                material = parts[1]

            if command == "f":
                # I guess this part of the code should be compatable
                # with POLYGON as well but IDK.
                face = []  # type: List[List[int]]
                for i in range(len(parts)):
                    if parts[i] in ["f", ""]:
                        continue
                    subs = parts[i].split("/")
                    vertex = int(subs[0]) - 1
                    textcoord = int(subs[1]) - 1 if len(subs) > 1 and len(subs[1]) > 0 else -1
                    normal = int(subs[2]) - 1 if len(subs) > 2 and len(subs[2]) > 0 else -1
                    face.append([vertex, textcoord, normal])
                if len(face) > 3:
                    logging.error("Only triangular wavefronts are accepted")
                    return

                _indices.append(face)
                _indice_materials.append(material)

        # Now unpack indices to actual object data
        i = 0
        fix_normals = False

        for _k, index in enumerate(_indices):
            ind = []
            for f in index:
                l_vertex = _vertices[f[0]]
                if f[2] != -1:
                    l_normal = _normals[f[2]]
                else:
                    fix_normals = True
                    l_normal = [0.0, 0.0, 1.0]
                l_tex = [0.0, 0.0]
                if f[1] > -1:
                    l_tex = _texcoords[f[1]]
                self._vertices.append(l_vertex)
                self._normals.append(l_normal)
                self._texcoords.append(l_tex)
                ind.append(i)
                i += 1
            self._indices.append(ind)
            self.materials[_indice_materials[_k]]._indices.append(ind)

        if fix_normals:
            self.fix_normals()


def export(mesh: Mesh, filename: str, name: str = "object") -> None:
    """
    Export given Mesh object as a Wavefront file with materials.

    Keyword arguments:
    filename -- Filename / path to load
    name -- Object name defined in exported file
    """
    if not isinstance(mesh, Mesh):
        logging.exception("Object is not an instance of Mesh")
        return None
    mat_filename = filename.replace(".obj", "") + ".mtl"

    output = [
        "# Payton Wavefront OBJ Exporter",
        f"mtllib {mat_filename}",
        f"o {name}",
    ]
    for v in mesh._vertices:
        output.append("v {}".format(" ".join(str(x) for x in v)))

    for t in mesh._texcoords:
        output.append("vt {}".format(" ".join(str(x) for x in t)))

    for n in mesh._normals:
        output.append("vn {}".format(" ".join(str(x) for x in n)))

    len_texcoords = len(mesh._texcoords) + 1
    len_normals = len(mesh._normals) + 1

    material_data = ["# Payton Wavefront OBJ Expoerter"]
    for name in mesh.materials:
        if len(mesh.materials[name]._indices) == 0:
            continue
        output.append(f"usemtl {name}")
        output.append("s off")
        material = mesh.materials[name]
        for indice in material._indices:
            f = [x + 1 for x in indice]
            t0 = str(f[0]) if len_texcoords > f[0] else ""
            n0 = str(f[0]) if len_normals > f[0] else ""

            t1 = str(f[1]) if len_texcoords > f[1] else ""
            n1 = str(f[1]) if len_normals > f[1] else ""

            t2 = str(f[2]) if len_texcoords > f[2] else ""
            n2 = str(f[2]) if len_normals > f[2] else ""
            output.append(f"f {f[0]}/{t0}/{n0} {f[1]}/{t1}/{n1} {f[2]}/{t2}/{n2}")
        material_data.append(f"newmtl {name}")
        material_data.append(f"Kd {material.color[0]} {material.color[1]} {material.color[2]}")
        if material.texture != "":
            base_name = os.path.basename(material.texture)
            path = os.path.join(os.path.abspath(os.path.dirname(filename)), base_name)
            copyfile(material.texture, path)
            material_data.append(f"map_Kd {base_name}")

    open(filename, "w").write("\n".join(output))
    open(mat_filename, "w").write("\n".join(material_data))



# https://gist.github.com/vladignatyev/06860ec2040cb497f0f3
# MIT Licensed
# Copyright (c) 2016 Vladimir Ignatev
import sys


def progress(count: int, total: int, status: str = '') -> None:
    bar_len = 60
    filled_len = int(round(bar_len * count / float(total)))

    percents = round(100.0 * count / float(total), 1)
    bar = '=' * filled_len + '-' * (bar_len - filled_len)

    sys.stdout.write('[%s] %s%s ...%s\r' % (bar, percents, '%', status))
    sys.stdout.flush()
